/*
  This program is free software published under the
  terms of the GNU Lesser General Public License.
  http://www.gnu.org/copyleft/lesser.html    
  This program is distributed in the hope that it will
  be useful - WITHOUT ANY WARRANTY; without even the
  implied warranty of MERCHANTABILITY or FITNESS FOR A
  PARTICULAR PURPOSE.
*/ 


/*  xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
     source code: ES6 (ECMAScript 6, ECMAScript 2015, JavaScript 6)
    xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */

//  xxxxxxxxxxxxxxx           version: 1.0           xxxxxxxxxxxxxxxx     

//  xxxxxxxxxxxxxxx    global variable declarations  xxxxxxxxxxxxxxxx


// These variables below can be directly modified by the user on the options form: 
// Commented variables are not globally (window scope) defined 

  // parad : type of paradigm
	
var item_cnt = 1;       // number of items within the task
 
var dots = 26;          // number of dots within each item 

  // shape => Shape type: dot (circle) or square
  
	// after => After the correct selection, the shape becomes.. '1'..'3'

  // rst_fixed => The raster is variable or fixed

var back_clr= '#fff';  // background window color during task

var dot_clr= '#000';  // shape background color 

var font_clr= '#fff';  // font color during task
	   
var timer_enabled = false;   // if true => timer is enabled. if false timer is disabled. Exactly: the timer is always on

var timer = 600;     // Task time in seconds if clock = 1000ms (default).

   // demo = false;     // if (demo == true) =>  In demo mode, each subsequent correct selection will be different color

   // force_speed = false;  Force the speed of work

   // force_precise = false;  Force the most precise control of the mouse path between individual target points

var store = false;      // false => this web application will not store results within client web browser.; true => This web application will store the results inside client web browser. 

// Other global constants and variables:

const clock = 1000;  // basic timer resolution in ms

var timer_id = null;   // timer_id is the identifier of the repeated action you want to cancel. This ID is returned from setInterval().

var counter = 0;   // timer counter. if (counter >= time) Timer will call theEnd

var is_task = false;    // global flag: task in progress or not. is_task == true for all events between startTask() and theEnd() events

var dirty = false;  // if  (dirty==true) => user changed any options in form 

var raster = 10;       // canvas (viewport) raster size	

var item = 0;   // current item in task     

var is_storage= false; // local storage is enabled on user browser or not

var startTime;  // start date time for current item in task

var is_resize= false;  // pageResize() event raised during task. see theEnd(), drawStim()

var stim = [];   //   [shape number]  x [is visible or not, visible shape value, hidden shape value]. see function initItem()

var prevX, prevY;  // Previous mouse coordinates (x, y)

var trail;         // trail for current item. unit: pixels (px)

var itime= [];     // time for each item in task (s)

var ierror= [];   // numbers of errors in task

var itrail= [];    // mouse trail for each item

var ioptimal= [];  // optimal path for each item


/*  xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
    xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
		xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */


// application init [create/start Timer, check storage, show start panel..] 
function init() 
{    
  
	if (timer_id)  {   
    window.clearInterval(timer_id);  
  }  
  timer_id= setInterval(function () {appTimer()},clock);  
  
  is_storage= isStorage() && storageAvailable('localStorage');
  if (is_storage)  {
    document.getElementById('store_group').style.display= 'block';
		document.getElementById('no_storage').style.display= 'none';
  }		
		
	setIns();

	showStart();   
}


// application timer handler function 
function appTimer()  
{    
	if (timer_enabled && is_task) // else: nop, idle, do nothing 
	{  
		counter++;  
		if (counter >= timer)  {     // task time is limited
			theEnd();
			return;
		}	
	}  
}	


// before using web storage, check browser support for localStorage and sessionStorage
function isStorage()  
{
  if (typeof(Storage) !== 'undefined') {
     return true;                 
  } 
  else {
    return false;    
  }  
}


// necessarily because of IE11 (and Edge) offline mode
function storageAvailable(type) 
{
  try  {
    return (window[type]) ? true: false;   
  }
  catch(e) {
    return false;
  }
}


// raised when page is resized. see html: <body onresize="pageResize()"
function pageResize()  
{ 
	if (is_task) 
	{ 
    is_resize= true;
		let c= document.getElementById('canvas');
		//c.style.display= 'block';
		c.width= window.innerWidth;  
		c.style.width= canvas.width + 'px';	     
		c.height= window.innerHeight;   
		c.style.height= canvas.height + 'px';  // https://css-tricks.com/whats-the-difference-between-width-height-in-css-and-width-height-html-attributes/ 		 		
		drawStim();	  // repaint stimulus		
	}
}	


// show start panel 
function showStart()  
{      
  document.getElementById('end').style.display= 'none';
  document.getElementById('result_wrapper').style.display= 'none';  
  document.getElementById('store_id').style.display= 'none';
	document.getElementById('table_grp').style.display= 'none';
	document.getElementById('start').style.display= 'block';	
}


// set instructions text depending on user selected task options
function setIns()  
{
  document.getElementById('instructions').style.display= 'block';		       
  
	document.getElementById('ins_shape').textContent= document.getElementById('shape').value == '1' ? 'Circles (dots) will soon appear on the screen.': 'Squares (dots) will soon appear on the screen.';

	let p= document.getElementById('parad');
	document.getElementById('ins_parad').textContent= p.options[p.selectedIndex].getAttribute('data-help');
	let a =  document.getElementById('after').value;
	if ( a == '1' ) {
		document.getElementById('ins_after').textContent= 'Once you click the dot correctly it will disappear.';		
	}
	else  {
		document.getElementById('ins_after').textContent= 'Once you click the dot correctly, it will no longer display textual content.';
		if ( a == '3' ) {
		  document.getElementById('ins_after').textContent += ' A trace will be automatically drawn between successfully completed dots.'; 	
		}
	}
	
	if (item_cnt == 1) {
		document.getElementById('ins_items').textContent= 'The task is completed when you have selected all the dots on the screen in the correct order.'; 
	}
	else  {
		document.getElementById('ins_items').textContent= 'After you finish the first task, the next task will be displayed automatically.';
	}
	
	document.getElementById('ins_timer').style.display= timer_enabled ? 'block' : 'none';	    // Warning: operating time is limited. 'The duration of this task is limited.'  
	
	document.getElementById('ins_demo').style.display= document.getElementById('demo').checked ? 'block' : 'none';	
	
	document.getElementById('ins_force_speed').style.display= document.getElementById('force_speed').checked ? 'block' : 'none';
	
	document.getElementById('ins_force_precise').style.display= document.getElementById('force_precise').checked ? 'block' : 'none';
	
}	


// refresh Options (update *_msg label) => give user short option HELP
function refOpt(e)  
{   
  dirty= true;       
  let my_id = e.id, msg = document.getElementById(my_id + '_msg'), s = 'Values within this range are allowed: ', s2= ' with step= ';               
	switch (my_id)
	{
		
		case 'parad':
		  msg.textContent= e.options[e.selectedIndex].getAttribute('data-help');		  						
			let ind= e.selectedIndex, my_max= 100;  // ind = 0 => Numbers-ascending; 1 = Numbers-descending
			if ( ind == 2 || ind == 3 )  {  // ind = 2 => Letters-ascending; 3 == Letters-descending
				my_max= 26;  // my_max= array length of letters A..Z = 26. 
			}		
			if ( ind == 4 || ind == 5 )  {  // ind = 4 => Numbers-letters ascending; 5 => Letters-numbers ascending
				my_max= 52;  // my_max= 2 * (array length of letters A..Z = 26.) 
			}										
			document.getElementById('dots').max= my_max;  // update dots maximum as function of selected paradigm type
			if (!document.getElementById('dots').validity.valid) {
				document.getElementById('dots_msg').textContent= 'Values within this range are allowed: 2..' + my_max + ' with step= 2';  // update dots validator
			  document.getElementById('dots_msg').setAttribute('class', 'error'); 
			}
			else  {
				document.getElementById('dots_msg').textContent= '';
        document.getElementById('dots_msg').removeAttribute('class'); 				
			}
			break;
		
		case 'item_cnt':	
			if (e.validity.valid)  {           
				msg.removeAttribute('class');       
				msg.textContent=  e.value == 1 ? 'After completing the first item, the tasks end.' : 'When the current task (item) is completed, the next task (item) is automatically displayed.'; 
			}
			else  {
				msg.textContent= s + e.min + '..' + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');    // msg.className = 'error';
			}      
			break;
		
		case 'dots':	
			if (e.validity.valid)  {           
				msg.removeAttribute('class');       
				msg.textContent=  ''; 
			}
			else  {
				msg.textContent= s + e.min + '..' + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');    // msg.className = 'error';
			}      
      break;
					
		case 'after':
		  switch (e.value) 
			{
				case '1':
					msg.textContent=  'After the correct selection, shape (dot) becomes invisible.';
					break;
        case '2':				
					msg.textContent= 'After selecting correctly, the shape (dot) is displayed without a digit (letter). If the user clicks on such an inactive dot, the error will not be counted.';
					break;
				case '3':				
					msg.textContent= 'After the correct selection of a minimum of two dots, a trail will be drawn between the correctly selected dots. After selecting correctly, the shape (dot) is displayed without a digit (letter). If the user clicks on such an inactive dot, the error will not be counted.';
					break;	
			}
		 break;
		
		
    case 'rst_fixed':
		  msg.textContent= e.value == '2' ? 'The size of the dots (and font) is inversely proportional to the number of dots used.' : 'The size of the dots (and font) does not depend on the number of dots displayed, but only on the screen size.'; 
			break;

    case 'back_clr': 
		case 'dot_clr':
		case 'font_clr':
			if (validColor(e.value))   {           
				msg.removeAttribute('class');       
				msg.textContent= '';  
			}
			else  {
				msg.textContent= 'The selected color is invalid.';
				msg.setAttribute('class', 'error');    // msg.className = 'error';
			} 
			break; 
			
    case 'timer_enabled':   
      if (e.checked)  {
        msg.textContent= 'Yes';
        document.getElementById('p_timer').style.display= 'block';
      }
      else  {  
        msg.textContent= 'Timer is disabled';
        document.getElementById('p_timer').style.display= 'none'; 
      }       
      break;
		
    case 'timer':
			if (e.validity.valid)  {
				msg.textContent= stringTime(e.value); 
				msg.removeAttribute('class');
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');
			} 
			break;  

		case 'demo': 
			msg.textContent= e.checked ? 'On. Each subsequent point you have to select will be displayed in red.' : 'Off';    
			break;
	
    case 'force_speed':
		case 'force_precise':
			msg.textContent= e.checked ? 'On' : 'Off';    
			break;
	
    case 'store':                                                                                                                                            
			msg.textContent= (e.value === '1') ? 'This web application will store the results inside your web browser (no expiration date). Useful for group research' : 'This web application will not store results within your web browser. Data is lost when the browser tab is closed';
			break;
 
	}   // switch
	
  refMsg();   
}


// refresh main validator message for user defined OPTIONS 
function refMsg()   
{ 	
	if ( optionsValid() )  {
    document.getElementById('is_valid').textContent=  dirty == true ? 'All your settings have been successfully updated' : '';  
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#000000'); 
  }
  else  {
    document.getElementById('is_valid').textContent='Some of your options are incorrectly set. They were not updated.';   // These settings are marked in red. 
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#ff0000;font-weight:bold'); 
  } 	
}


// check for validity.valid object in all input type="number". see: setPar() and autoCorrect()
function optionsValid()  
{ 
  let res= document.getElementById('item_cnt').validity.valid && document.getElementById('dots').validity.valid && document.getElementById('timer').validity.valid;
  return res;              
}


// return integer into string hh:mm:ss time format
// input par: total == integer (time in seconds)
function stringTime(total) 
{
  let hh = Math.floor(total/(60*60));         
  let div_min = total % (60 * 60);
  let mm = Math.floor(div_min / 60);       
  let div_sec = div_min % 60;
  let ss = Math.ceil(div_sec);    
  // These lines ensure you always have two-digits. This formats your string to HH:MM:SS
  if (hh < 10)  {hh = "0"+hh;}     
  if (mm < 10) {mm = "0"+mm;}
  if (ss < 10) {ss = "0"+ss;} 
  return hh + ":" + mm + ":" + ss;
} 


// start task (research) 
function startTask(e)   
{     
  	
  initLayout();     			
  
	setPar();
			
	drawStim();
  	
  is_task= true;    // finally "activate" timer   
	      
}


// init visible/invisible html dom elements, progress bar and background colors
function initLayout()
{   	 
	document.getElementById('result_wrapper').style.display= 'none';
  document.getElementById('end').style.display= 'none'; 
  document.getElementById('start').style.display= 'none'; 
}	


// set internal application research parameters (global vars)
function setPar()  
{  
  // this section is probably redundant (danger of indefinite loops)
	if (dots > 100)  { 
		dots= 100; 
	}	
	let ind= document.getElementById('parad').selectedIndex;  // ind = 0 => Numbers-ascending; 1 = Numbers-descending
	if ( dots > 26 && (ind == 2 || ind == 3) )  {  // 2 => Letters-ascending; 3 == Letters-descending
   	dots= 26;  // max= array length of letters A..Z = 26. 
  }		
	if ( dots > 52 && (ind == 4 || ind == 5) )  {  // 4 => Numbers-letters ascending; 5 => Letters-numbers ascending
   	dots= 52;  // max= 2 * (array length of letters A..Z = 26.) 
  }	
	
	is_resize= false;
  itime= [];  		// time for each item in task 
	ierror= [];   	// numbers of errors in task
	itrail= [];   	// mouse trail for each item
	ioptimal= [];   // optimal path for each item
	item= 0; 	
	let rst_fixed = document.getElementById('rst_fixed').value == '1' ? true : false;  // '1' => Fixed or '2' => Variable; 
  raster = rst_fixed ? 10 : Math.ceil(Math.sqrt(dots));   // raster is fixed (true) or variable (false)
	
	initItem();
	
  counter= 0;       // reset timer counter
}


// init and generate current item. stimulus generator
function initItem()
{
  
	let c= document.getElementById('canvas');
	c.style.display= 'block';
	c.width= window.innerWidth;  
	c.style.width= canvas.width + 'px';	     
	c.height= window.innerHeight;   
	c.style.height= canvas.height + 'px';  
  let dx = c.width / raster, dy = c.height / raster;
	let path= [];   // tmp array of x,y coord of all visible shapes for ioptimal array
					
	const UP = raster * raster - 1; 		// high bound of array
  let letters= [];   // array length= 26. Upper case letters A..Z. for lower case use: for (let j= 97; j<=122; j++) {  lowerCase letters 97..122. array length= 26. 
  for (let j= 65; j<=90; j++) { // upper case 65..90. length= 26
    letters.push(String.fromCharCode(j));  // var str= String.fromCharCode(ascii_code);
	}		
  
	// reset global var for current item 
	stim= [];	   // reset current stimulus
	trail=0;     // reset mouse path for current item
  prevX= null; // previous mouse coordinates (x, y)
  prevY= null;
	
	for (let i= 0; i <= UP; i++)  {     
		stim[i]= [0,null,null];	
	}
  // now change random selected dots..
	let	sel={}, j, ind;	// sel => already selected cells within current item. j= selected random index 
	// init start value of index (for visible state of shape - index 1)
	
	let parad= document.getElementById('parad').value;   // paradigm type value (see html tag)
	switch (parad)   { 	
		case '2':
		case '4':
      ind= dots;  // descending  
			break;	
		default:
		  ind= 0;     // ascending
	}
	
	for (let i= 1; i <= dots; i++)    // i => number of visible dot (shapes) within current item 
	{		
		do  {	
			j= getRndInt(0,UP);
		} while ( typeof sel[j] != 'undefined');
		sel[j]= 1;  // remember this position (shape, dot)		
			// index 0: this dot (shape) will be initially visible (0 = invisible; 1 = visible)
			// index 1: visible shape value. this value is shown to user 
			// index 2: hidden shape value. see isCorrect()	
		switch (parad) 
		{
			case '1':   // The order of correct selection is: 1-2-3..
			  stim[j]= [ 1, i, i ];   	// stim[j][0]= 1;  stim[j][1]= i; stim[j][2]= i;
				break;
			
			case '2':  // The order of correct selection is: ..3-2-1
			  stim[j]= [ 1, ind, i ];
				ind--;
				break;
			
			case '3':   // The order of correct selection is: A-B-C..
			  stim[j]= [ 1, letters[i-1], i ];   
				break;
			
			case '4':   // The order of correct selection is: ..C-B-A
			  stim[j]= [ 1, letters[ind-1], i ];   
				ind--;
				break;			
			
			case '5':  // The order of correct selection is: 1-A-2-B-3-C..
			  if ( i % 2 == 0 ) {
					stim[j]= [ 1, letters[ind], i ];  // letters in ascending order A,B,C..
					ind++;
				}
				else  {
					stim[j]= [ 1, ind+1, i ];  // numbers in ascending order: 1,2,3,..
				}
				break;
			
			case '6':  // The order of correct selection is: A-1-B-2-C-3..
			  if ( i % 2 != 0 ) {
					stim[j]= [ 1, letters[ind], i ];  // letters in ascending order A,B,C..
					ind++;
				}
				else  {
					stim[j]= [ 1, ind, i ];  // numbers in ascending order: 1,2,3,..
				}
				break;
						
		}
		let row = Math.floor (j / raster);
		let col = Math.floor (j % raster);  // or => col= pos - row * raster;  // pos = row * raster + col;               						
		let x= col*dx+dx/2, y= row*dy+dy/2;
		path.push([x,y]);	
	}
	let dist= 0, len= path.length;
	for (let i=1; i< len; i++)  {
	  dist+= getDistance(path[i-1][0], path[i-1][1], path[i][0], path[i][1]);   // function getDistance(x1, y1, x2, y2)  {
	}
	ioptimal[item]= dist;  // compute optimal path here
  ierror[item]= 0;       // init number of errors in current item
	startTime = new Date();	
	
}	


// get random integer from min to max range
function getRndInt(min, max) {
	return Math.floor(Math.random() * (max - min + 1)) + min;
}


// Changes application settings that can be changed directly by the user within the options form
function setVarPar()  
{   
  	  			
  if (document.getElementById('item_cnt').validity.valid) {
		item_cnt= parseInt(document.getElementById('item_cnt').value, 10); 
	} 
  
	if (document.getElementById('dots').validity.valid) {
		dots= parseInt(document.getElementById('dots').value, 10); 
	} 		  
 		
  if (validColor(document.getElementById('back_clr').value))  {		
		back_clr= document.getElementById('back_clr').value;
  }
	
	if (validColor(document.getElementById('dot_clr').value))  {		
		dot_clr= document.getElementById('dot_clr').value;
  }
	
  if (validColor(document.getElementById('font_clr').value))  {		
		font_clr= document.getElementById('font_clr').value;
  }	
	
	
	timer_enabled= (document.getElementById('timer_enabled').checked) ? true: false;  // timer is enabled or disabled
	
	if (timer_enabled && document.getElementById('timer').validity.valid) {
		timer= Math.round(parseInt(document.getElementById('timer').value, 10));  // task time (s)
  }	
							  			     
  store= (is_storage && document.getElementById('store').value === '1') ? true : false;
	
}


// test color for validity (from string)
// https://stackoverflow.com/questions/8027423/how-to-check-if-a-string-is-a-valid-hex-color-representation/8027444
function validColor(clr){
  return /^#[0-9A-F]{6}$/i.test(clr);
}


// draw stimulus (all dots for current item)         
function drawStim() 
{    		
  let c= document.getElementById('canvas');	
	let ctx= c.getContext('2d');  
	ctx.clearRect(0, 0, c.clientWidth, c.clientHeight);  // clear canvas object

	ctx.fillStyle= back_clr;
	ctx.fillRect(0, 0, c.clientWidth, c.clientHeight);	
		
	let dx = c.width / raster, dy = c.height / raster;
	let r = Math.floor(Math.min(dx, dy)/2);  // this is circle (dot or square) radius (line length). font-size
	
	let d_clr= dot_clr, f_clr= font_clr, first= -1, done=[];
	let is_dot= document.getElementById('shape').value == 1 ? true : false; 
	// After the correct selection, the shape becomes: '1' == Invisible; '2' == No number (letter); '3' == Show Trail   
	let after= document.getElementById('after').value;  
	let demo = document.getElementById('demo').checked;
	if (demo)  {
		first= getFirst();
	}	
	let len = stim.length;
	for (let j= 0; j < len; j++)
	{	    	     
		if (stim[j][1] == null) { 
			continue;  // see initItem()
		}		
		let row = Math.floor (j / raster);
		let col = Math.floor (j % raster);  // or => col= pos - row * raster;  // pos = row * raster + col;               						
		let x= col*dx+dx/2, y= row*dy+dy/2, txt = stim[j][1];	 
		// draw shape (circle or square)
		if (stim[j][0] == 1 || after != '1') 
		{			
			ctx.fillStyle= (j == first && demo) ? '#f00' : d_clr;	 		
			ctx.beginPath(); 			
			if (is_dot)  {				  
				ctx.arc(x, y, r, 0, 2*Math.PI);  //ctx.arc(x,y,r,sAngle,eAngle,counterclockwise);                 				
			}
			else  {
				ctx.rect(x-r, y-r, 2*r, 2*r);   							
			}
			ctx.fill();
		}
		// draw text on shape
		if (stim[j][0] == 1)
		{
			ctx.fillStyle= f_clr;
			ctx.font = r.toString() + 'px Arial';   
         //ctx.font.fontFamily = 'Arial, Verdana, Helvetica, sans-serif';  // https://www.w3schools.com/cssref/css_websafe_fonts.asp
			ctx.textAlign = 'center';
			ctx.textBaseline= 'middle';							
			ctx.fillText(txt, x, y);
		}
    else if (after == '3') { 
			done.push([x,y,stim[j][2]]);
		}		
	} 
	
  // connect the finished dots if is_trail == true 
  if (after == '3' && done.length > 1)  
	{ 
    //done.sort((a,b) => a[2] - b[2]);  // this not works in IE11
		done.sort(function(a, b) {
			return a[2] - b[2];
		});				
		ctx.strokeStyle = d_clr; 		
		ctx.beginPath();		
		ctx.moveTo(done[0][0], done[0][1]);	
		for (let i= 1; i< done.length; i++) {	
			ctx.lineTo(done[i][0], done[i][1]); 
		}
    ctx.stroke();			
	}
		
}


// return position for first non-selected shape
// if all are selected return -1. use this function in demo mode
function getFirst()
{
	const UP = raster * raster;   
	let min= 1000, res= -1; 
	for (let i= 0; i < UP; i++)  {     
		if ( stim[i][0] == 1 && stim[i][2] < min ) {
			 min= stim[i][2];
			 res= i;
		}	
	}
	return res;			
}


// goto next item or end task
function nextItem()  
{ 
		
	// see:  https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Date/getTime
  // convert time difference from milliseconds to seconds
	let endTime = new Date(); 
  itime[item]= ( endTime.getTime() - startTime.getTime() ) / 1000;  // seconds  (item == current item in task)	
	itrail[item]= trail;   // save user mouse path (trail) for current item
	
	item++;			
		
	if (item < item_cnt)  {         		 
     initItem();
     drawStim();		 
  }
  else  {                
		theEnd();  // end of task if item >= item_cnt  
  }
}


// end of task (research)
function theEnd()  
{   

  is_task= false;    	
	
	document.getElementById('canvas').style.display= 'none'; 
  document.getElementById('end').style.display= 'block';   
  document.getElementById('result_wrapper').style.display= 'block'; 
	
  const D = 2;  // number of decimal	
  // points = sum of scores for each item in task. For each item: if the item was successfully completed = 1 point. Else = 0 point.   
	let points = 0, t= 0, redun= [], rows = []; // t = total task time 
	let header= is_resize ? ['Task', 'Point', 'Errors', 'Time (s)'] : ['Task', 'Point', 'Errors', 'Time (s)', 'Your Trail', 'Optimal Trail', 'Trail Redundancy'];   
	let errors= 0; // errors = number of errors in task. 
	for (let i in ierror) {
		errors+= ierror[i];
	}
	for (let i= 0; i< itime.length; i++)  // see condition: i< itime.length; 
	{   // table1 rows are printed only for completed items => see function nextItem()   		 
		points++;		
		t+= itime[i];		// t => total task time (sum of times for all items)
		if (is_resize) {	
			rows[i]= [ parseInt(i+1,10), 1, ierror[i], itime[i].toFixed(D) ];  			
		}
		else  
		{
			redun.push( itrail[i] / ioptimal[i] );		
			rows[i]= [ parseInt(i+1,10), 1, ierror[i], itime[i].toFixed(D), itrail[i].toFixed(D), ioptimal[i].toFixed(D), ( itrail[i] / ioptimal[i] ).toFixed(D) ];  			
		}			 
	}
  let mis = item_cnt - points;	// mis => number of missing items in task.
	if ( points > 0 )  {
		rows.push( [ 'Sum', points,  errors, t.toFixed(D), '', '', '' ] ); 
	}
     
	createTableArray(document.getElementById('table1'), rows, header);	
	
	// calculate decriptive statistics => table2  
	let n, m, sd;
	rows= [];
	let up = is_resize ? 2 : 3;
	for (let j= 0; j< up; j++)
	{
		let arr= [], s= '';
		if ( j == 0 ) {
		  arr= ierror;  // 2. row => items errors 
      s= 'Errors';			
		}
		else if ( j == 1 ) {
		  arr= itime;  // 2. row => items time  
			s= 'Time';
		}
		else {
		  arr= redun;  // 2. row => path redundancy => user trail / optimal trail  
			s= 'Redundancy';
		}
		n= arr.length;
		m = (n > 0) ? mean(arr).toFixed(D) : '';     
		sd = (n > 1) ? stDev(arr, m).toFixed(D): ''; 	
		rows.push( [s, m, sd, n] );  // last row
	}
	let tmp= ['Variable', 'Average', 'Sample Standard Deviation', 'N']; // header
	createTableArray(document.getElementById('table2'), rows, tmp); 
	
	document.getElementById('score').textContent= 'Sum of points= ' + points + '. Maximum points= ' + item_cnt + '. Total errors= ' + errors + '.  Missing= ' + mis + '.'; 
	
	document.getElementById('l_ts').textContent='Last completed research: ' + new Date();
  document.getElementById('l_res').style.display= 'block';
	
	try 
  {
		if ( isNaN(t) || t <= 0 ) {
			t= '';
		}
		else  {
			t= t.toFixed(D);
		}
    if (store) 
		{
			if (is_resize) {
				setStorage(points, errors, t, '', ''); 
			}
			else  {
				setStorage(points, errors, t, m, sd);
			}
    }
  }        
  catch (e)  { 
    document.getElementById('table_grp').style.display= 'block';
		document.getElementById('table_grp').textContent= "Your browser probably doesn't have support for the Web Storage Object or access to the Storage Object is disabled. Check your browser settings. In some browsers (IE11, Edge) Storage Object is not enabled when accessed in a local file (offline)."; 
  }    	 

}


// return arithmetic mean (mean, average) of array (arr) 
function mean(arr)  
{	
	let len= arr.length, sum= 0;
	for (let i=0; i<len; i++) {
		sum+= parseFloat(arr[i]);
	}
	return (sum/arr.length);
}	


// return Sample Standard Deviation (with Bessel's correction N-1)
// arr = array argument; mean = array average 
function stDev(arr, mean) 
{
	let len= arr.length, sum= 0, res;
	for (let i=0; i<len; i++) {
	  sum+= Math.pow(arr[i]-mean, 2);	 // Math.pow(base, exponent)
	}
	res= Math.pow(sum/(len-1), 0.5);   // denominator = N-1
	return res;
}


// Formats the date in the desired format. If you do not want this string format do not call this function. use: new Date()
function setDate()  
{
	let today = new Date();
	let dd = today.getDate();
	let mm = today.getMonth()+1; 
	let yyyy = today.getFullYear();
	//if(dd<10) { dd='0'+dd; } 
	if(mm<10) {	mm='0'+mm; } 
	today = dd+'.' + mm + '.' +yyyy;
	return today;
}
 
 
/* Store the results of the last completed task in the user's browser
   set user data into window.localStorage. Parameters:	 
	 points = sum of points in task. each item => 0 or 1 point
   errors = sum of errors in task 
	 time => sum of time for all items in task in seconds 
	 m_red => mean for ratio:  user (mouse) trail / optimal trail 
   sd_red => sample standard deviation for ratio:  user (mouse) trail / optimal trail 	 
*/ 
function setStorage(points, errors, time, m_red, sd_red)   
{  
	let oldItems = JSON.parse(localStorage.getItem('dft-v1.0')) || []; 
  let id= oldItems.length+1;     	
	// set new "record"
  let newItem = {
      'id' : id,
      'date' : setDate(),   //new Date(), 			
      'items' : item_cnt,			
			'points' : points, 	
			'errors' : errors,
      'time-sum': time,
      'redundancy-m': (typeof m_red === 'undefined') ? '' : m_red,
			'redundancy-sd':(typeof sd_red === 'undefined') ? '': sd_red,			
			'demo' : document.getElementById('demo').checked ? 1 : 0,
			'paradigm' : document.getElementById('parad').value,
			'shape-count': dots,
			'shape-type': document.getElementById('shape').value,
			'after-select': document.getElementById('after').value,
			'raster': document.getElementById('rst_fixed').value,
			'timer-time': timer_enabled ? document.getElementById('timer').value : '',
      'force-speed': document.getElementById('force_speed').checked ? 1 : 0,
      'force-trail': document.getElementById('force_precise').checked ? 1 : 0 			
  };

  oldItems.push(newItem);

  localStorage.setItem('dft-v1.0', JSON.stringify(oldItems));
  
	document.getElementById('table_grp').style.display= 'block';
  document.getElementById('store_id').style.display= 'block';
  document.getElementById('store_id').textContent= 'The results are stored in your browser. ID= ' + id ;
  
	createTableJSON(document.getElementById('table_grp'), oldItems); 

}


// convert JSON data dynamically to HTML Table using JavaScript - Dynamically Create table with rows in JavaScript
// downloaded and adapted from source: https://www.encodedna.com/javascript/populate-json-data-to-html-table-using-javascript.htm
function createTableJSON(elem, my_json) 
{  
  // EXTRACT VALUE FOR HTML HEADER. 
  let col = [];
  for (let i = 0; i < my_json.length; i++) {
    for (let key in my_json[i]) {
        if (col.indexOf(key) === -1) {
            col.push(key);
        }
    }
  } 
  let table = document.createElement("table");
  // CREATE HTML TABLE HEADER ROW USING THE EXTRACTED HEADERS ABOVE.
  let tr = table.insertRow(-1);                   // TABLE ROW.
  for (let i = 0; i < col.length; i++) {
      let th = document.createElement("th");      // TABLE HEADER.
      th.innerHTML = col[i];
      tr.appendChild(th);
  }
  // ADD JSON DATA TO THE TABLE AS ROWS.
  for (let i = 0; i < my_json.length; i++) 
  {
    tr = table.insertRow(-1);
    for (let j = 0; j < col.length; j++) {
      let tabCell = tr.insertCell(-1);
      tabCell.innerHTML = my_json[i][col[j]];
    }
  }
  // FINALLY ADD THE NEWLY CREATED TABLE WITH JSON DATA TO A CONTAINER.
  elem.innerHTML= "";
  elem.appendChild(table);
}


/* convert array data dynamically to HTML table - dynamically create html table
   elem == DOM html element (node)
   data => 1-dim or 2-dim array (matrix)
   col => string column descriptions. Set an empty string if you want to skip a column inside a 2-dim array	 */ 
function createTableArray(elem, data, col) 
{  
	let table= document.createElement("table");  // html table
  let tr= table.insertRow(-1), th, cell, tmp;  // table row, table header, table cell 	
  // create html table header row using the extracted headers above  
  for (let i in col) {
		if ( col[i] == '' ) { continue; } // set an empty string if you want to skip a column inside a 2-dim array
		th= document.createElement("th");      
		th.innerHTML = col[i];
		tr.appendChild(th);
  }
  if ( is2dArray(data) ) 
	{ // add 2-dim array data to the table as rows		
		tmp= data.length;
		for (let i = 0; i < tmp; i++) 
		{
			tr= table.insertRow(-1);
			for (let j in col) {
				if (col[j] == '') { continue; }  // set an empty string if you want to skip a column inside a 2-dim array
				cell= tr.insertCell(-1);
				cell.innerHTML = data[i][j];
			}
		}
  }
	else 
	{	// add 1-dim array data to the table as only one row
		tr= table.insertRow(-1);
		tmp=data.length;
		for (let i = 0; i < tmp; i++)	{
			cell= tr.insertCell(-1);
			cell.innerHTML = data[i];
		}
	}	
  elem.innerHTML= '';
  elem.appendChild(table); // finally add the newly created table with string data to a html container
}


// test if array is 2-dim (return true) or not (return false)
function is2dArray(arr)
{
	if(arr[0] === undefined){
		return false;
	}
	else	{
		return (arr[0].constructor === Array);
	}
}


// this onclick event is fired by html canvas element => document.getElementById('canvas')
function doAnswer(e)
{ 
  let canvas =  e.target || e.srcElement;  // document.getElementById('canvas');      	
	let pos = getCellFromXY(e.offsetX, e.offsetY);

	if ( typeof(pos) === 'undefined' )  {
		return;  // error
	}	
  else  {
		pos= parseInt(pos, 10);
	}	
  if ( stim[pos][0] != 1 )  {  // this is not error
    return; // abort all if circle (square) is not visible in this rectangle (cell)
	}	
	
	prevX= e.clientX;  // init x coord
	prevY= e.clientY;  // init y coord
  if (isCorrect(stim[pos][2])) {
		stim[pos][0]= 0;   // this shape now will be invisible  
	}
	else  {
		ierror[item]++;
	}
	
	drawStim();  // repaint
	
	if ( itemDone() )  {
		nextItem();
	}	
}


/* if all the visible dots on the screen there is none with a lower ordinal number returns true
	 if all the visible dots in the item there is one dot with a smaller ordinal number than n => the function returns false  
	 n => hidden dot ordinal number  seeInitItem() function */
function isCorrect(n)
{
	const UP = raster * raster;   
	for (let i= 0; i < UP; i++)  {     
		if ( stim[i][0] == 1 && stim[i][2] < n ) {
			return false;
		}	
	}
	return true;	
}


// check if current item is done (all shapes must be invisible => stim[i][0] != 1)
function itemDone() 
{
	const UP = raster * raster;   
	for (let i= 0; i < UP; i++)  {     
		if ( stim[i][0] == 1 ) {
			return false;
		}	
	}	
	return true;
}


// https://stackoverflow.com/questions/24926028/drag-and-drop-multiple-objects-in-html5-canvas   
// canvas => mousemove event	=> a pointing device is moved over an element (fired continuously as the mouse moves).
// this event is fired by html canvas element => document.getElementById('canvas')  
function mouseMove(e)
{  
  if ( prevX != null && !is_resize && is_task) 
	{
		try  {
			let d = getDistance(prevX, prevY, e.clientX, e.clientY);   //function getDistance(x1, y1, x2, y2) 
			trail+= d;
			prevX= e.clientX;  // save the current mouse position (x)
			prevY= e.clientY;  // save the current mouse position (y)
    } catch(e) {  // small probability of range error => overflow
			is_resize=true;
		}		
  } 
}


// return distance from point P1(x1,y1) to point P2(x2,y2)
function getDistance(x1, y1, x2, y2)  {
  return Math.sqrt( Math.pow(x1-x2, 2) + Math.pow(y1-y2, 2) ); // total line distance (px) 
}	


/* Get the cell (rectangle) ordinal position in html canvas element from x and y mouse position
   offsetX:  x - coordinate of the mouse pointer, relative to the target element. 
   offsetY:	 y - coordinate of the mouse pointer, relative to the target element. */
function getCellFromXY(offsetX, offsetY)
{	
  let canvas= document.getElementById('canvas');
	let dx =  canvas.width / raster, dy = canvas.height / raster;  
  let row = Math.floor ( offsetY / dy );  // The offsetX (offsetY) property returns the x (y) -coordinate of the mouse pointer, relative to the target element. 
  let col = Math.floor ( offsetX / dx );  // The clientX (clientY) property returns the horizontal (vertical) coordinate (according to the client area) of the mouse pointer when a mouse event was triggered.
  if (row > raster - 1 || row < 0 || col > raster - 1 || col < 0) {
    return;  // return 'undefined' value	=> abort all     
  }
  let pos = row * raster + col;   // rectangle position within html canvas object  	
	if ( pos >= raster * raster || pos < 0 ) {    
		return; // return 'undefined' value	=> abort all  
  }
	else {		
		return pos;	
	}
}	

 
// begin of new code for export localStorage app data:  data= localStorage.getItem('myApp-v1.0');

 
// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
// save to window global vars all user Options/Settings
function hideOption()  
{
  setVarPar();   
  document.getElementById('options').style.display = 'none'; 
  document.getElementById('start').setAttribute('class', 'center-xy');   // ex 'child' class 
  document.getElementById('show_option').style.display = 'block';   
  document.getElementById('start_div').style.display = 'block';        //.removeAttribute("disabled");  
  document.getElementById('hide_option').style.display = 'none';   
  document.getElementById('instructions').style.display= 'block';
  refMsg(); 
  setIns();	
}


// show form with application options
// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
function showOption()   
{
  document.getElementById('hide_option').style.display= 'block';
  document.getElementById('options').style.display= 'block';
  document.getElementById('start').classList.remove('center-xy');  // ex 'child' class    
  document.getElementById('start_div').style.display= 'none';    
  document.getElementById('show_option').style.display= 'none';  
  document.getElementById('instructions').style.display= 'none';     
  document.getElementById('store_oper').style.display= 'none';
	if (document.getElementById('store_group').disabled === false && localStorage.getItem('dft-v1.0')) 
  {
    let cnt= JSON.parse(localStorage.getItem('dft-v1.0')).length;                                                        
    document.getElementById('store_cnt').textContent=  'You have a total of ' + cnt  + ' results stored in your web browser.';
    if (cnt > 0) {
			document.getElementById('store_oper').style.display= 'table';   // xxx
		}
	}
  dirty= false;
  document.getElementById('is_valid').textContent='';  // tmp reset validator message
}


// clear web browser storage data for this application
function clrStorage()  
{  
  let cnt= localStorage.getItem('dft-v1.0') ? JSON.parse(localStorage.getItem('dft-v1.0')).length : 0;                                                        
  if (cnt === 0)  {
    alert('You have no data to delete');
    return;
  }  
  if ( !window.confirm('You have a total of ' + cnt  + ' results stored in your web browser. Continue?') ) { 
   return;
 }                                             
 if (window.confirm('Delete total saved application data:  '  + cnt + '. This action cannot be undone. Final confirmation?') ) 
 {  
   window.localStorage.removeItem('dft-v1.0'); 
   document.getElementById('store_cnt').textContent= 'All data has been deleted. No data.';
	 document.getElementById('store_oper').style.display= 'none';
 }  
}


// copy data from user web browser to the Clipboard 
function copyStorage()  
{  
  let data= localStorage.getItem('dft-v1.0');
	if (!data)  {
    alert('You have no data to Copy');
    return;
  }  
	data = JSON.parse(data); 
	let arr= jsonToArray(data);	
	copy2DToClipboard(arr);
}	


// convert json object into 2D array
// data = JSON object
// return 2D array
function jsonToArray(data)  
{
   // extract column (fields) names  
  let total= data.length, col = [], row, res=[];  // res= result (2d array) 
  for (let i = 0; i < total; i++) {
    for (let key in data[i]) {
			if (col.indexOf(key) === -1)  {
				col.push(key);  // col= columns (fields) names
			}
    }
  } 
	res.push(col);  // add field names as 1 row
  let c = col.length;    // column count  
  // add JSON data to the 2d array as rows.
  for (let i = 0; i < total; i++) 
  {
		row= [];
    for (let j = 0; j < c; j++)  {     
      row.push(data[i][col[j]]);
    }
		res.push(row);
  }
	return res;	
}


// https://stackoverflow.com/questions/60698115/how-to-copy-a-javascript-2d-array-to-clipboard-to-paste-it-in-excel
// copy 2D array to Clipboard
function copy2DToClipboard(arr) {
  let csv = '', row, cell;
  for (row = 0; row < arr.length; row++) {
    for (cell = 0; cell < arr[row].length; cell++) {
      csv += (arr[row][cell]+'').replace(/[\n\t]+/g, ' ');
      if (cell+1 < arr[row].length) csv += '\t';
    }
    if (row+1 < arr.length) { 
			csv += '\n'; 
		}
  }
  copyTextToClipboard(csv);
}


// copy 2D array (or text) to Clipboard
function copyTextToClipboard(txt) 
{	
	if (!navigator.clipboard) {
    fallbackCopy(txt);
    return;
  }
  navigator.clipboard.writeText(txt).then(function() {
   //alert('Async: Copying to Clipboard was successful!');
  }, function(err) {
    alert('Oops, could not copy data. Error: ' + err);
  });
}


// copied from https://stackoverflow.com/questions/400212/how-do-i-copy-to-the-clipboard-in-javascript
// fallback copy text to clipboard
function fallbackCopy(txt) 
{
  let ta = document.createElement('textarea');
  ta.value = txt; 
  // Avoid scrolling to bottom
  ta.style.top = '0';
  ta.style.left = '0';
  ta.style.position = 'fixed';
  document.body.appendChild(ta);
  ta.focus();
  ta.select();
  try {
    let successful = document.execCommand('copy');
    let msg = successful ? 'successful' : 'unsuccessful';
  } catch (err) {
    alert('Oops, unable to copy. Error: ' + err);
  }
  document.body.removeChild(ta);
}


// https://stackoverflow.com/questions/2004555/how-can-i-make-javascript-make-produce-new-page
// displays user saved results (Window.localStorage application data) in new empty web page (tab) 
function newWin()
{ 
	let data = JSON.parse(localStorage.getItem('dft-v1.0'))
	createTableJSON(document.getElementById('table_grp'), data);
	let stl= '<style>table{margin:1em auto;border:thick solid #000;border-collapse:collapse}td,th{border:medium solid #000;text-align:center;padding:.5em}th{font-weight:bold}</style>';
	let tbl= document.getElementById('table_grp').innerHTML;	
	let win = window.open('');
  win.document.write('<html><head><title>Saved cumulative data</title>' + stl + '</head><body>' + tbl + '</body></html>');
  win.focus();
}


// https://stackoverflow.com/questions/2004555/how-can-i-make-javascript-make-produce-new-page
// displays Latest research results in new empty web page (tab) 
// parameter: id => DOM/html object id (div or table)
function newWinLast(id)
{ 
  let ind = document.getElementById('last_data_sel').selectedIndex;
	ind++;
	let e = document.getElementById('table' + ind);		
	let stl= '<style>table{margin:1em auto;border:thick solid #000;border-collapse:collapse}td,th{border:medium solid #000;text-align:center;padding:.5em}th{font-weight:bold}</style>';
	let tbl= e.innerHTML; 	
	let win = window.open('');
	win.document.write('<html><head><title>Latest research data</title>' + stl + '</head><body>' + tbl + '</body></html>');
	win.focus();
}


// copy only latest task results (data) to the Clipboard 
function copyLast()  
{     
  let ind = document.getElementById('last_data_sel').selectedIndex, tbl;
  switch (ind) {
    case 0:
			tbl = document.getElementById('table1').getElementsByTagName('table')[0];  // note: id='table1' is html div element
			break;		
		default:
			tbl = document.getElementById('table2').getElementsByTagName('table')[0];  // note: id='table2' is html div element
			//break;	
	}		
	
	if (tbl)  {
		let data= table2DArray(tbl);
		if (data.length > 0) {
			copy2DToClipboard(data);
		}
	}
}	


// https://stackoverflow.com/questions/3065342/how-do-i-iterate-through-table-rows-and-cells-in-javascript
// input argument tbl == html/DOM  table  object
// return 2D array
function table2DArray(tbl) 
{
	let res = [];	 
	for (let i = 0, row; row = tbl.rows[i]; i++) 
	{   //iterate through rows. rows would be accessed using the "row" variable assigned in the for loop
	  res[i]= [];		
		for (let j = 0, col; col = row.cells[j]; j++) {  //iterate through columns. columns would be accessed using the "col" variable assigned in the for loop			 
		  res[i][j]= col.textContent;
		}  
	}
	return res;				
}
 